/*
 * Copyright (C) 2003-2017 FAUmachine Team <info@faumachine.org>.
 * This program is free software. You can redistribute it and/or modify it
 * under the terms of the GNU General Public License, either version 2 of
 * the License, or (at your option) any later version. See COPYING.
 */

#include "config.h"

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include "glue.h"
#include "umutil.h"

struct cpssp {
	unsigned int state_power;
	unsigned char data[SZ_BIOS_ROM];
};

static void *
CHIP_(access)(struct cpssp *cpssp, unsigned long pa)
{
	assert(/* 0 <= pa && */ pa < SZ_BIOS_ROM);

	pa &= 0xfffff000;

	return &cpssp->data[pa];
}

static int
CHIP_(readb)(void *_cpssp, uint32_t addr, uint8_t *valp)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;
	uint8_t *va;

	addr &= SZ_BIOS_ROM - 1;

	assert(/* 0 <= pa && */ addr < SZ_BIOS_ROM);

	/*
	 * Read Contents
	 */
	va = CHIP_(access)(cpssp, addr & 0xfffff000);
	*valp = *(va + (addr & 0x00000fff));

	return 0;
}

static int
CHIP_(writeb)(void *_cpssp, uint32_t addr, uint8_t val)
{
	// struct cpssp *cpssp = (struct cpssp *) _cpssp;

	return 0;
}

static int
CHIP_(map)(
	void *_cpssp,
	uint32_t addr,
	char **haddr_p
)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	addr &= SZ_BIOS_ROM - 1;

	assert(/* 0 <= addr && */ addr < SZ_BIOS_ROM);

	*haddr_p = (char *) &cpssp->data[addr];

	return 0;
}

static int
CHIP_(map_r)(
	void *_cpssp,
	uint32_t addr,
	int (**cfp)(void *, uint32_t, unsigned int, uint32_t *),
	void **csp,
	char **haddr_p
)
{
	if (CHIP_(map)(_cpssp, addr, haddr_p) == 0) {
		*cfp = NULL;
		*csp = NULL;
		return 0;
	}
	return 1;
}

static int
CHIP_(map_w)(
	void *_cpssp,
	uint32_t addr,
	int (**cfp)(void *, uint32_t, unsigned int, uint32_t),
	void **csp,
	char **haddr_p
)
{
	if (CHIP_(map)(_cpssp, addr, haddr_p) == 0) {
		*cfp = NULL;
		*csp = NULL;
		return 0;
	}
	return 1;
}

static void
CHIP_(power_set)(void *_cpssp, unsigned int val)
{
	struct cpssp *cpssp = (struct cpssp *) _cpssp;

	cpssp->state_power = val;
	/* FIXME */
}

void *
CHIP_(create)(
	const char *name,
	const char *img,
	struct sig_manage *port_manage,
	struct sig_std_logic *port_power,
	struct sig_cs *port_cs,
	struct sig_mem_bus *port_bus
)
{
	static const struct sig_std_logic_funcs power_funcs = {
		.boolean_or_set = CHIP_(power_set),
	};
	static const struct sig_cs_funcs cs_funcs = {
		.readb = CHIP_(readb),
		.writeb = CHIP_(writeb),
		.map_r = CHIP_(map_r),
		.map_w = CHIP_(map_w),
	};
	struct cpssp *cpssp;
	const char *path;
	int fd;
	int ret;

	cpssp = shm_alloc(sizeof(*cpssp));
	assert(cpssp);

	path = buildpath(ROMDIR, img);
	fd = open(path, O_RDONLY);
	assert(0 <= fd);

	ret = read(fd, cpssp->data, SZ_BIOS_ROM);
	if (ret < SZ_BIOS_ROM) {
		fprintf(stderr, "WARNING: %s: %s: too small.\n",
				CHIP, img);
	} else {
		char c;

		ret = read(fd, &c, 1);
		if (0 < ret) {
			fprintf(stderr, "WARNING: %s: %s: too big.\n",
					CHIP, img);
		}
	}

	ret = close(fd);
	assert(0 <= ret);

	/* Out */
	/* Call */
	sig_cs_connect(port_cs, cpssp, &cs_funcs);

	/* In */
	cpssp->state_power = 0;
	sig_std_logic_connect_in(port_power, cpssp, &power_funcs);

	return cpssp;
}

void
CHIP_(destroy)(void *_cpssp)
{
	struct cpssp *cpssp = _cpssp;

	shm_free(cpssp);
}

void
CHIP_(suspend)(void *_cpssp, FILE *fp)
{
	struct cpssp *cpssp = _cpssp;
	
	generic_suspend(cpssp, sizeof(*cpssp), fp);
}

void
CHIP_(resume)(void *_cpssp, FILE *fp)
{
	struct cpssp *cpssp = _cpssp;
	
	generic_resume(cpssp, sizeof(*cpssp), fp);
}
